require 'active_record'

module KarrierWave
  module Mount

    def kage_options
      @kage_options ||= superclass.respond_to?(:kage_options) ? superclass.kage_options.dup : {}
    end

    def mount_kage(column, options = {})
      include KarrierWave::Mount::Extension

      before_save :"store_#{column}!"
      before_save :"write_#{column}_meta"
      after_commit :"remove_#{column}", on: :destroy

      kage_options[column.to_sym] = options

      class_eval <<-RUBY, __FILE__, __LINE__+1
        serialize :#{options[:mount_on] || column}

        def #{column}
          _mounter(:#{column})
        end

        def #{column}=(new_file)
          _mounter(:#{column}).cache(new_file)
        end

        def remote_#{column}_url
          _mounter(:#{column}).remote_url
        end

        def remote_#{column}_url=(url)
          _mounter(:#{column}).remote_url = url
        end

        def remote_#{column}_object_id
          _mounter(:#{column}).remote_object_id
        end

        def remote_#{column}_object_id=(object_id)
          _mounter(:#{column}).remote_object_id = object_id
        end

        def #{column}?
          _mounter(:#{column}).present?
        end

        def store_#{column}!
          _mounter(:#{column}).store!
        end

        def write_#{column}_meta
          _mounter(:#{column}).write_meta
        end

        def remove_#{column}
          _mounter(:#{column}).remove!
        end
      RUBY
    end

    module Extension
      private

      def _mounter(column)
        @_mounters ||= {}
        @_mounters[column] ||= Mounter.new(self, column)
      end
    end

    class Mounter
      attr_reader :column, :record, :remote_url, :remote_object_id

      def initialize(record, column, options = {})
        @record = record
        @column = column
        @options = record.class.kage_options[column]
      end

      def url(version = nil)
        kage_file.url(version_keys[version]) || default_url(version)
      end

      def snapshot_url(version = nil)
        kage_file.snapshot_url(version_keys[version]) || default_url(version)
      end

      def default_url(version = nil)
        if @options[:default_url].is_a? String
          ActionController::Base.helpers.asset_path(@options[:default_url])
        elsif @options[:default_url].is_a? Array
          url = @options[:default_url][record.id % @options[:default_url].length]
          ActionController::Base.helpers.asset_path(url)
        elsif @options[:default_url].is_a? Hash
          url = @options[:default_url][version || :default]
          ActionController::Base.helpers.asset_path(url)
        end
      end

      def write_meta
        will_update? && record.send(:write_attribute, serialization_column, kage_file.to_json)
      end

      def cache(new_file)
        if new_file
          @file = KarrierWave::LocalFile.new(new_file)
        else
          @remove = true
        end

        will_update!
      end

      def kage_file
        @kage_file ||= KarrierWave::File.new(self, @options[:config_name],
          identifier: identifier,
          filename: filename,
          snapshot_filename: @options[:snapshot_filename],
          type: type
        )
      end

      def identifier
        read_meta[:key]
      end

      def type
        @options[:type] || read_meta[:type]
      end

      def filename
        @options[:filename] || read_meta[:filename]
      end

      def width
        read_meta[:width]
      end

      def height
        read_meta[:height]
      end

      def read_meta
        record.read_attribute(serialization_column).to_h || {}
      end

      def serialization_column
        @options[:mount_on] || column
      end

      def version_keys
        @options[:version_keys] || {}
      end

      def store!
        if remove?
          kage_file.delete
        else
          !@file.blank? && kage_file.store(@file)
        end

        will_update!
      end

      def remove?
        @remove
      end

      def remove!
        kage_file.delete
        will_update!
      end

      def remote_url=(url)
        @remote_url = url
        @file = KarrierWave::RemoteFile.new(URI.parse(url).open(proxy: KarrierWave.config.http_proxy))
        will_update!
      end

      def remote_object_id=(object_id)
        @remote_object_id = object_id
        @file = KarrierWave::RemoteObject.new(object_id)
        will_update!
      end

      protected

      def will_update!
        record.send(:"#{serialization_column}_will_change!")
        @meta_updated = true
      end

      def will_update?
        record.send(:"#{serialization_column}_changed?") && @meta_updated
      end
    end
  end
end

ActiveRecord::Base.extend KarrierWave::Mount
